import java.io.IOException;
import java.io.PrintStream;

import java.lang.InterruptedException;
import java.lang.Thread;

import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Scanner;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;

import java.util.logging.Logger;
import java.util.logging.FileHandler;
import java.util.logging.ConsoleHandler;
import java.util.logging.Formatter;
import java.util.logging.SimpleFormatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.LogManager;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.Statement;
import java.sql.SQLException;

import misc.*;
import network.*;
import game.*;
import game.attack.*;
import game.defense.*;

class Simulator {
  ExecutorService threadPool;
  PrintStream out = System.out;
  Logger log;
  Connection connection;
  
  Network networkCAIDA;
  
  Simulator(int nThreads) throws IOException, SQLException {    
    LogManager.getLogManager().reset();
    log = Logger.getLogger(System.class.getName());
    
    ConsoleHandler consoleHandler = new ConsoleHandler();
    Formatter formatter = new Formatter() {
      @Override
      public String format(LogRecord record) {
        return record.getMessage();
      }
    };
    consoleHandler.setFormatter(formatter);  
    consoleHandler.setLevel(Level.INFO);      
    log.addHandler(consoleHandler);
    
    FileHandler fileHandler = new FileHandler("../log/simulator.log");
    fileHandler.setFormatter(new SimpleFormatter());
    fileHandler.setLevel(Level.FINEST);
    log.addHandler(fileHandler);
    
    threadPool = Executors.newFixedThreadPool(nThreads); 
    connection = DriverManager.getConnection("jdbc:sqlite:../results/results.db");
    
    networkCAIDA = NetworkGenerators.loadCAIDA(null, 0.5);
  }
  
  void saveResult(HashMap<String,Object> result) throws SQLException {
    ArrayList<String> keys = new ArrayList<String>(result.keySet());
    ArrayList<String> values = new ArrayList<String>();
    for (String key : keys)
      values.add("\"" + result.get(key).toString() + "\"");
    
    synchronized (connection) {
      Statement statement = connection.createStatement();
      String sql = "REPLACE INTO results (" + String.join(", ", keys) + ") VALUES (" + String.join(", ", values) + ")";
      log.fine(sql);
      statement.executeUpdate(sql);
      statement.close();
    }
  }
  
  List<Experiment> waiting = Collections.synchronizedList(new ArrayList<Experiment>());
  List<Experiment> running = Collections.synchronizedList(new ArrayList<Experiment>());
  List<Experiment> finished = Collections.synchronizedList(new ArrayList<Experiment>());
  
  class Experiment implements Runnable {
    String setup;
    
    Experiment(String setup) {
      this.setup = setup;
    }
    
    HashMap<String,Object> performExperiment() {
      String[] parts = setup.split("/");    
      
      String setting = parts[0];
      String modelName = parts[1];
      String networkName = parts[2];
      String algorithm = parts[3]; 
      Integer budget = new Integer(parts[4]);
      
      Integer iterations = new Integer(parts[5]);
      Integer instances = new Integer(parts[6]);
      
      PropagationModel model = null;
      if (modelName.equals("independent"))
        model = IndependentCascades.getInstance();  
      else if (modelName.equals("repeated"))
        model = RepeatedIndependentCascades.getInstance(); 
        
      Double utility = 0.0;
      Double time = 0.0;
        
      for (int i = 0; i < instances; i++) {
        Network network = null;
        if (networkName.equals("CAIDA"))
          network = networkCAIDA;
        else if (networkName.equals("ER"))
          network = NetworkGenerators.generateErdosRenyiGraph(100, 0.05, 0.5, new Random(i));
        else if (networkName.equals("BA"))
          network = NetworkGenerators.generateBarabasiAlbert(100, 3, 3, 0.5, new Random(i));
          
        Random random = new Random(i);
        ArrayList<Network.Node> remainingNodes = new ArrayList<Network.Node>(network.getNodes());
        HashSet<Network.Node> targetSet = Subsets.randomSubset(remainingNodes, 1, random);
        remainingNodes.removeAll(targetSet);
        HashSet<Network.Node> initialBaseSet = Subsets.randomSubset(remainingNodes, 10, random);
        remainingNodes.removeAll(initialBaseSet);
        HashSet<Network.Node> monitoringBaseSet = Subsets.randomSubset(remainingNodes, 10, random);
        
        StackelbergGame game = new StackelbergGame(network, model, targetSet, monitoringBaseSet, initialBaseSet); 
        game.ITERATIONS = iterations;
        HashSet<Network.Node> monitoringSet = null;
        
        if (setting.equals("distributional")) {
          game.initialSetDistribution = true;
          
          long startTime = System.nanoTime();

          if (algorithm.equals("Optimal"))
            monitoringSet = OptimalDefenseResponse.getInstance().findDefenseResponse(game, budget, initialBaseSet);
          else if (algorithm.equals("Greedy3"))
            monitoringSet = Greedy3.getInstance().findDefenseResponse(game, budget, initialBaseSet);
          
          double estimatedTime = (System.nanoTime() - startTime) / 1000000000.0; 
          
          time += estimatedTime;
          utility += game.defenderPayoff(monitoringSet, initialBaseSet);          
        }
        else if (setting.equals("maximin")) {
          AttackStrategyAlgorithm attackAlgorithm = OptimalAttackStrategy.getInstance();
          int attackBudget = 1;
          
          long startTime = System.nanoTime();

          if (algorithm.equals("Optimal"))
            monitoringSet = OptimalDefenseStrategy.getInstance().findDefenseStrategy(game, budget, attackAlgorithm, attackBudget);
          else if (algorithm.equals("Greedy1"))
            monitoringSet = Greedy1Strategy.getInstance().findDefenseStrategy(game, budget, attackAlgorithm, attackBudget);
          else if (algorithm.equals("Greedy3"))
            monitoringSet = Greedy3Strategy.getInstance().findDefenseStrategy(game, budget, attackAlgorithm, attackBudget);
          else if (algorithm.equals("Greedy4"))
            monitoringSet = Greedy4Strategy.getInstance().findDefenseStrategy(game, budget, attackAlgorithm, attackBudget);
          
          double estimatedTime = (System.nanoTime() - startTime) / 1000000000.0; 
          
          HashSet<Network.Node> initialSet = attackAlgorithm.findBestResponse(game, monitoringSet, attackBudget);
          
          time += estimatedTime;
          utility += game.defenderPayoff(monitoringSet, initialSet);          
        }
      }
      
      HashMap<String,Object> result = new HashMap<String,Object>();
      result.put("setting", setting);
      result.put("model", modelName);
      result.put("network", networkName);
      result.put("algorithm", algorithm);
      result.put("budget", budget);
      result.put("iterations", iterations);
      result.put("instances", instances);
      result.put("utility", utility / instances);
      result.put("time", time / instances);
      
      return result;
    }
    
    public void run() {
      waiting.remove(this);
      running.add(this);
      
      try {
        log.fine("Starting experiment " + setup + "...");
        HashMap<String,Object> result = performExperiment();
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
        result.put("datetime", dateFormat.format(new Date()));
        saveResult(result);
        log.fine("Finished experiment " + setup + ".");
      }
      catch (SQLException e) {
        log.severe(e.toString());
      }
      
      running.remove(this);
      finished.add(this);
    }
    
    public String toString() {
      return setup;
    }
  }
  
  void printStatus() {
    out.printf("Finished (%d):\n", finished.size());
    synchronized(finished) {
      for (Experiment experiment : finished)
        out.printf("  %s\n", experiment.toString());
    }
    
    out.printf("Running (%d):\n", running.size());
    synchronized(waiting) {
      for (Experiment experiment : running)
        out.printf("  %s\n", experiment.toString());
    }
    
    out.printf("Waiting (%d):\n", waiting.size());
    synchronized(waiting) {
      for (Experiment experiment : waiting)
        out.printf("  %s\n", experiment.toString());
    }
  }
  
  void processCommand(String command) {
    try {
      String[] parts = command.trim().split(" ");    
      String cmd = parts[0];
      
      if (cmd.equals("quit") || cmd.equals("q")) {
        synchronized (connection) {
          connection.close();          
          System.exit(0);
        }
      }
      else if (cmd.equals("run") || cmd.equals("r")) {
        Experiment experiment = new Experiment(parts[1]);
        waiting.add(experiment);
        threadPool.execute(experiment);
      }
      else if (cmd.equals("status") || cmd.equals("s"))
        printStatus();
      else if (cmd.equals(""))
        ;
      else
        out.println("Uknown command!");
    }
    catch (Exception e) {
      out.println(e.toString());
    }
  }
  
  void run() {
    Scanner scanner = new Scanner(System.in);    
    while (true) {
      out.print(">");
      processCommand(scanner.nextLine());
    }
  }
  
  public static void main(String[] args) throws IOException, SQLException {
    Simulator simulator = new Simulator(7);
    simulator.run();
  }
}
