package network;

import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.Random;
import java.util.Arrays;
import java.util.Collections;

public class IndependentCascades extends PropagationModel {
  private static final IndependentCascades instance = new IndependentCascades();
  private IndependentCascades() {}
  public static IndependentCascades getInstance() { return instance; } 
  
  @Override
  protected boolean simulatePropagation(Network network, Set<Network.Node> initialSet, Set<Network.Node> monitoringSet, Set<Network.Node> targetSet, Random random) {
    System.out.println("This should not happen!");
    
    int numNodes = network.getNodes().size();
    boolean[] infected = new boolean[numNodes];
    int[] nodeType = new int[numNodes]; // 0: regular, 1: initial, 2: monitoring, 3: target

    int[] openList = new int[numNodes]; // always stores the nodes infected in the previous round
    int openListSize = 0;

    ArrayList<Integer> initialList = new ArrayList<Integer>();    
    for (Network.Node node : initialSet)
      initialList.add(node.ID);      
    Collections.sort(initialList);
    
    for (Integer nodeID : initialList) {
      openList[openListSize++] = nodeID;
      infected[nodeID] = true;
    }
    
    for (Network.Node node : monitoringSet)
      nodeType[node.ID] = MONITORING;      
    for (Network.Node node : targetSet)
      nodeType[node.ID] = TARGET;    
    
    int[] spread = new int[numNodes]; // will store the newly infected nodes     
    int spreadSize = 0; // ``clear'' the spread array

    while (true) {
      boolean detected = false;
      
      for (int i = 0; i < openListSize; i++) {
        int nodeID = openList[i]; 
        
          for (int outgoingArcID : network._arcs[nodeID]) {
            int neighborID = network._arcTo[outgoingArcID];
            if (!infected[neighborID])
              if (random.nextDouble() < network._arcWeight[outgoingArcID]) {
              if (nodeType[neighborID] == TARGET)
                return true;  // epidemic reached target node first
              if (nodeType[neighborID] == MONITORING)
                detected = true; // epidemic reaches monitoring node
                
              infected[neighborID] = true;
              spread[spreadSize++] = neighborID;
            }
        }
      }
                  
      if (detected)
        return false; // epidemic reached monitoring node
      if (spreadSize == 0)
        return false; // epidemic died

      int[] tmp = openList;
      openList = spread; // ``copy'' the spread array
      openListSize = spreadSize;
      spread = tmp;
      spreadSize = 0; // ``clear'' the spread array              
    }
  }
  
  @Override
  public double simulatePropagation(Network network, Set<Network.Node> initialSet, Set<Network.Node> monitoringSet, Set<Network.Node> targetSet, int iterations) {
    Random random;
    int numNodes = network.getNodes().size();
    boolean[] infected = new boolean[numNodes];
    int[] spread = new int[numNodes]; // stores the newly infected nodes in each round    
    int[] nodeType = new int[numNodes]; // 0: regular, 1: initial, 2: monitoring, 3: target
    int[] openList = new int[numNodes]; // stores the nodes infected in the previous round

    for (Network.Node node : monitoringSet)
      nodeType[node.ID] = MONITORING;      
    for (Network.Node node : targetSet)
      nodeType[node.ID] = TARGET;    
     
    ArrayList<Integer> initialList = new ArrayList<Integer>();    
    for (Network.Node node : initialSet)
      initialList.add(node.ID);      
    Collections.sort(initialList);
      
    int successes = 0;
      
    ITERATION:
    for (int iteration = 0; iteration < iterations; iteration++) {
      random = new Random(iteration);
      
      for (int nodeID = 0; nodeID < numNodes; nodeID++)
        infected[nodeID] = false;
    
      int openListSize = 0;

      for (Integer nodeID : initialList) {
        openList[openListSize++] = nodeID;
        infected[nodeID] = true;
      }
    
      int spreadSize = 0; // ``clear'' the spread array

      while (true) {
        boolean detected = false;
        
        for (int i = 0; i < openListSize; i++) {
          int nodeID = openList[i]; 
          
          for (int outgoingArcID : network._arcs[nodeID]) {
            int neighborID = network._arcTo[outgoingArcID];
            if (!infected[neighborID])
              if (random.nextDouble() < network._arcWeight[outgoingArcID]) {
                if (nodeType[neighborID] == TARGET) {
                  successes++;  // epidemic reached target node first
                  continue ITERATION;
                }
                
                if (nodeType[neighborID] == MONITORING)
                  detected = true; // epidemic reaches monitoring node
                  
                infected[neighborID] = true;
                spread[spreadSize++] = neighborID;
              }
          }
        }
                    
        if (detected)
          continue ITERATION; // epidemic reached monitoring node
        if (spreadSize == 0)
          continue ITERATION; // epidemic died

        int[] tmp = openList;
        openList = spread; // ``copy'' the spread array
        openListSize = spreadSize;
        spread = tmp;
        spreadSize = 0; // ``clear'' the spread array              
      }
    }
    
    return successes / (double)iterations;
  }
}

