package network;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Collections;

import game.StackelbergGame;

public abstract class NetworkGenerators {  
  public static Network generateErdosRenyiGraph(int size, double probability, double weight, Random random) {    
    Network network = new Network();
    
    for (int i = 0; i < size; i++) {
      Network.Node node = network.addNode();
      
      for (int j = 0; j < i; j++)
        if (random.nextDouble() < probability) {
          Network.Node otherNode = network.getNode(j);
          network.addEdge(node, otherNode, weight);
        }
    }
    
    network.createArrays();
    
    return network;
  }
  
  public static Network generateBarabasiAlbert(int size, int clique, double degree, double weight, Random random) { 
    assert clique >= degree;
    
    Network network = new Network();
    
    for (int i = 0; i < size; i++)
      network.addNode();
    
    ArrayList<Network.Node> popularity = new ArrayList<Network.Node>();
    
    // create inital clique
    for (int i = 0; i < clique; i++) {
      Network.Node node = network.getNode(i);
      for (int j = 0; j < i; j++) {
        Network.Node otherNode = network.getNode(j);
        network.addEdge(node, otherNode, weight);        
        popularity.add(node);
        popularity.add(otherNode);
      }
    }
    
    // add additional nodes
    for (int i = clique; i < size; i++) {
      Network.Node node = network.getNode(i);
      HashSet<Network.Node> neighbors = new HashSet<Network.Node>();
      while (neighbors.size() < degree)
        neighbors.add(popularity.get(Math.abs(random.nextInt()) % popularity.size()));
        
      ArrayList<Network.Node> neighborList = new ArrayList<Network.Node>(neighbors);
      Collections.sort(neighborList);
      
      for (Network.Node neighbor : neighborList) { 
        network.addEdge(node, neighbor, weight);
        popularity.add(node);
        popularity.add(neighbor);
      }
    }
    
    network.createArrays();
    
    return network;
  }
  
  public static StackelbergGame generateTricky(PropagationModel model) {
    double epsilon = 0.001;
    
    Network network = new Network();
    
    HashSet<Network.Node> targetSet = new HashSet<Network.Node>();
    HashSet<Network.Node> initialBaseSet = new HashSet<Network.Node>();
    HashSet<Network.Node> monitoringBaseSet = new HashSet<Network.Node>();
    
    Network.Node target = network.addNode();
    targetSet.add(target);
    
    Network.Node monitor = network.addNode();
    monitoringBaseSet.add(monitor);
    
    Network.Node a = network.addNode();
    initialBaseSet.add(a);
    
    ArrayList<Network.Node> b = new ArrayList<Network.Node>();
    for (int i = 0; i < 10; i++) {
      Network.Node bi = network.addNode();
      initialBaseSet.add(bi);
      b.add(bi);
    }

    Network.Node i1 = network.addNode();
    Network.Node i2 = network.addNode();
    Network.Node aux = network.addNode();
    
    network.addArc(i1, monitor, 0.55);
    network.addArc(i2, aux, 1.0);
    network.addArc(aux, target, 1.0);
    
    network.addArc(a, i2, epsilon);
    
    for (Network.Node bi : b) {
      network.addArc(bi, i2, epsilon);
      network.addArc(bi, i1, 1.0);
    }
    
    network.createArrays();

    return new StackelbergGame(network, model, targetSet, monitoringBaseSet, initialBaseSet);
  }
  
//  public static Network generateSetCover(int baseSet, int subsets, int subsetSize, double weight, HashSet<Network.Node> targetSet, HashSet<Network.Node> initialBaseSet, HashSet<Network.Node> monitoringBaseSet, Random random) {
//    targetSet.removeAll(targetSet); //TODO: fixme! should rather return a game
//    initialBaseSet.removeAll(initialBaseSet);
//    monitoringBaseSet.removeAll(monitoringBaseSet);
//        
//    Network network = new Network();
//    
//    Network.Node target = network.addNode();
//    targetSet.add(target);
//    
//    Network.Node intermediate = network.addNode();
//    network.addArc(intermediate, target, 1.0);
//    
//    ArrayList<Network.Node> elements = new ArrayList<Network.Node>();
//    for (int i = 0; i < baseSet; i++) {
//      Network.Node element = network.addNode();
//      initialBaseSet.add(element);
//      network.addArc(element, intermediate, weight);
//      elements.add(element);
//    }
//    
//    for (int i = 0; i < subsets; i++) {
//      Network.Node subset = network.addNode();
//      monitoringBaseSet.add(subset);
//      ArrayList<Network.Node> neighbors = new ArrayList<Network.Node>();
//      while (neighbors.size() < subsetSize)
//        neighbors.add(elements.get(Math.abs(random.nextInt()) % elements.size()));
//      for (Network.Node element : neighbors) 
//        network.addEdge(element, subset, weight);
//    }
//    
//    network.createArrays();
//    
//    return network;
//  }
  
  public static Network loadCAIDA(String country, double weight) throws IOException {   
    Network network = new Network();
    
    HashMap<String,String> orgIDtoCountry = new HashMap<String,String>();
    
    // read as-org2info
    BufferedReader reader = new BufferedReader(new FileReader("../data/CAIDA/20150101.as-org2info.txt"));
    String line = "#";
    
    // skip beginning
    while (line.startsWith("#")) {
      line = reader.readLine();
    }
    
    // process organizations
    while (!line.startsWith("#")) {
      String[] fields = line.split("\\|");
      orgIDtoCountry.put(fields[0], fields[3]);
      line = reader.readLine();
    }
    
    // skip middle
    // skip beginning
    while (line.startsWith("#")) {
      line = reader.readLine();
    }  
    
    HashMap<String,String> asIDtoCountry = new HashMap<String,String>();

    // process systems
    while (line != null) {
      String[] fields = line.split("\\|");
      asIDtoCountry.put(fields[0], orgIDtoCountry.get(fields[3]));
      line = reader.readLine();
    }    
              
    // read as-rel
    reader = new BufferedReader(new FileReader("../data/CAIDA/20150101.as-rel.txt"));
    line = "#";
    
    // skip beginning
    while (line.startsWith("#")) {
      line = reader.readLine();
    }

    HashSet<String> filtered = new HashSet<String>();
    for (HashMap.Entry<String, String> entry : asIDtoCountry.entrySet())
      if ((country == null) || (entry.getValue().equals(country)))
        filtered.add(entry.getKey());
        
    HashSet<String> connected = new HashSet<String>();

    // preprocess relations
    while (line != null) {
      String[] fields = line.split("\\|");
      if (filtered.contains(fields[0]) && filtered.contains(fields[1])) {
        connected.add(fields[0]);
        connected.add(fields[1]);
      }
      line = reader.readLine();
    }     
    
    HashMap<String,Network.Node> asIDtoNode = new HashMap<String,Network.Node>();
    for (String ID : connected)
      asIDtoNode.put(ID, network.addNode());
    
    // read as-rel again
    reader = new BufferedReader(new FileReader("../data/CAIDA/20150101.as-rel.txt"));
    line = "#";
    
    // skip beginning
    while (line.startsWith("#")) {
      line = reader.readLine();
    }

    // process relations
    while (line != null) {
      String[] fields = line.split("\\|");
      String from = fields[0];
      String to = fields[1];
      if (connected.contains(from) && connected.contains(to))
        network.addEdge(asIDtoNode.get(from), asIDtoNode.get(to), weight);
      line = reader.readLine();
    }     
    
    network.createArrays();
    return network;
  }
}
